#!/usr/bin/env python

"""
Confluence Wiki XML/XHTML syntax parsing.

Copyright (C) 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from common import *
from xmlread import Parser
import re
import sys
import operator
import htmlentitydefs
import codecs

# XML dialect syntax parsing.

tags = {
    # XHTML tag               MoinMoin syntax
    "strong"                : "'''%s'''",
    "em"                    : "''%s''",
    "u"                     : "__%s__",
    "del"                   : "--(%s)--",
    "sup"                   : "^%s^",
    "sub"                   : ",,%s,,",
    "code"                  : "`%s`",
    "pre"                   : "{{{%s}}}",
    "table"                 : "{{{#!table\n%s\n}}}",
    "tbody"                 : "%s",
    "tr"                    : "%s",
    "th"                    : "'''%s'''",
    "td"                    : "%s",
    "blockquote"            : " %s",
    "small"                 : "~-%s-~",
    "big"                   : "~+%s+~",
    "p"                     : "%s",
    "ol"                    : "%s",
    "ul"                    : "%s",
    "ac:plain-text-body"    : "{{{%s}}}",
    "ac:link"               : "[[%s%s|%s]]",
    "ac:image"              : "{{%s%s|%s}}",
    }

for tag, translation in blocktypes.items():
    tags[tag] = translation

simple_tags = {
    # XHTML tag               MoinMoin syntax
    "br"                    : "<<BR>>",
    }

list_tags = {
    # XHTML list tag          MoinMoin list item syntax
    "ol"                    : "1. %s",
    "ul"                    : "* %s",
    }

indented_tags = ["li", "p"]

link_target_tags = {
    # Confluence element      Attribute providing the target
    "ri:page"               : "ri:content-title",
    "ri:attachment"         : "ri:filename",
    "ri:user"               : "ri:username",
    }

macro_rich_text_styles = {
    # Confluence style        MoinMoin admonition style
    "note"                  : "caution",
    "warning"               : "warning",
    "info"                  : "important",
    "tip"                   : "tip",
    }

normalise_regexp_str = r"\s+"
normalise_regexp = re.compile(normalise_regexp_str)

class ConfluenceXMLParser(Parser):

    "Handle content from Confluence 4 page revisions."

    def __init__(self, out):
        Parser.__init__(self)
        self.out = out

        # Link target information.

        self.target = None
        self.target_type = None

        # Macro information.

        self.macro = None
        self.macro_parameters = {}

        # Indentation and preformatted states.

        self.indent = 0
        self.states = {}
        for name in ("pre", "ac:plain-text-body"):
            self.states[name] = 0

        # Table states.

        self.table_rows = 0
        self.table_columns = 0

    # ContentHandler-related methods.

    def startElement(self, name, attrs):
        if list_tags.has_key(name):
            self.indent += 1
        elif self.states.has_key(name):
            self.states[name] += 1
        Parser.startElement(self, name, attrs)

    def endElement(self, name):
        Parser.endElement(self, name)
        if list_tags.has_key(name):
            self.indent -= 1
        elif self.states.has_key(name):
            self.states[name] -= 1

    def characters(self, content):
        if not self.is_preformatted():
            content = self.normalise(content, self.elements[-1])
        Parser.characters(self, content)

    def skippedEntity(self, name):
        ch = htmlentitydefs.name2codepoint.get(name)
        if ch:
            self.text[-1].append(unichr(ch))

    # Parser-related methods.

    def handleElement(self, name):
        text = "".join(self.text[-1])

        # Handle state.

        if name == "table":
            self.table_rows = 0
        elif name == "tr":
            self.table_columns = 0

        # Find conversions.

        conversion = None

        # Handle list elements.

        if name == "li" and len(self.elements) > 1:
            list_tag = self.elements[-2]
            conversion = list_tags.get(list_tag)

        # Remember link target information.

        elif link_target_tags.has_key(name):
            self.target = self.attributes[-1].get(link_target_tags[name])
            self.target_type = name
            text = ""

        # Remember macro information.

        elif name == "ac:parameter":
            self.macro_parameters[self.attributes[-1].get("ac:name")] = text
            text = ""

        elif name == "ac:macro":
            self.macro = self.attributes[-1].get("ac:name")

        # Handle the common case.

        else:
            conversion = tags.get(name)

        # Attempt to convert the text.

        # Links require target information.
        # NOTE: User links should support the intended user namespace prefix.

        if name in ("ac:link", "ac:image"):
            if self.target_type == "ri:attachment":
                prefix = "attachment:"
            elif self.target_type == "ri:user":
                prefix = ""
            else:
                prefix = "../"

            text = conversion % (prefix, self.target, text or self.target)
            self.target = self.target_type = None

        # Macro name information is used to style rich text body regions.

        elif name == "ac:macro" and macro_rich_text_styles.has_key(self.macro):
            details = macro_rich_text_styles[self.macro]
            title = self.macro_parameters.get("title")
            if title:
                details = "%s\n\n%s" % (details, title)
            text = "{{{#!wiki %s\n\n%s}}}" % (details, text)
            self.macro = None
            self.macro_parameters = {}

        # Handle the common case.

        elif text and conversion:
            text = conversion % text
        elif simple_tags.has_key(name):
            text = simple_tags[name]

        # Postprocess table columns and rows.

        if name in ("th", "td"):
            if self.table_columns:
                text = "\n|| %s" % text
            self.table_columns += 1
        elif name == "tr":
            if self.table_rows:
                text = "\n==\n%s" % text
            self.table_rows += 1

        # Normalise leading whitespace and indent the text if appropriate.

        if name in indented_tags:
            text = " " * self.indent + text.lstrip()

        # Add the converted text to the end of the parent element's text nodes.

        if len(self.text) > 1:
            nodes = self.text[-2]
            if "".join(self.text[-2]):
                parent = self.elements[-2]
                if parent == "body":
                    nodes.append("\n\n")
                elif list_tags.has_key(parent):
                    nodes.append("\n")
                elif list_tags.has_key(name):
                    nodes.append("\n")
            nodes.append(text)

        # Otherwise, emit the text.

        else:
            self.out.write(text)

    def is_preformatted(self):
        return reduce(operator.or_, self.states.values(), False)

    # Whitespace normalisation.

    def get_replacement(self, name):
        if name in ("html", "body", "table", "tbody", "tr") or list_tags.has_key(name):
            return ""
        else:
            return " "

    def normalise(self, text, name):
        return normalise_regexp.sub(self.get_replacement(name), text)

def parse(s, out):

    "Parse the content in the string 's', writing a translation to 'out'."

    # NOTE: CDATA sections appear to have erroneous endings.

    s = u"""\
<?xml version="1.0"?>
<!DOCTYPE html 
     PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
     "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<body>
%s
</body>
</html>""" % s.replace("]] >", "]]>")

    f = StringIO(s.encode("utf-8"))
    try:
        parser = ConfluenceXMLParser(out)
        parser.parse(f)
    finally:
        f.close()

if __name__ == "__main__":
    s = sys.stdin.read()
    out = codecs.getwriter("utf-8")(sys.stdout)
    parse(s, out)

# vim: tabstop=4 expandtab shiftwidth=4
